@@ -1,4 +1,6 @@
 #  Copyright (c) 2020. Lena "Teekeks" During <info@teawork.de>
+"""Helper functions"""
+
 import urllib.parse
 import uuid
 from typing import Union, List, Type, Optional
@@ -6,12 +8,25 @@ from json import JSONDecodeError
 from aiohttp.web import Request
 from dateutil import parser as du_parser
 from enum import Enum
+from .types import AuthScope
+from urllib.parse import urlparse, parse_qs
 
 
 TWITCH_API_BASE_URL = "https://api.twitch.tv/helix/"
 TWITCH_AUTH_BASE_URL = "https://id.twitch.tv/"
 
 
+def extract_uuid_str_from_url(url: str) -> Union[str, None]:
+    """Extracts a UUID string from a URL
+
+    :param str url: The URL to parse
+    :return: UUID string extracted from given URL or None if no UUID found
+    :rtype: Union[str, None]
+    """
+    uuids = parse_qs(urlparse(url).query).get('uuid', [])
+    return uuids[0] if len(uuids) > 0 else None
+
+
 def build_url(url: str, params: dict, remove_none=False, split_lists=False) -> str:
     """Build a valid url string
 
@@ -66,9 +81,9 @@ async def get_json(request: 'Request') -> Union[list, dict, None]:
 def make_fields_datetime(data: Union[dict, list], fields: List[str]):
     """Itterates over dict or list recursivly to replace string fields with datetime
 
-    :param data: dict or list
-    :param fields: list of keys to be replaced
-    :rtype: dict or list
+    :param union[dict, list] data: dict or list
+    :param list[str] fields: list of keys to be replaced
+    :rtype: union[dict, list]
     """
 
     def make_str_field_datetime(data, fields: list):
@@ -99,11 +114,12 @@ def make_fields_datetime(data: Union[dict, list], fields: List[str]):
         return make_str_field_datetime(data, fields)
 
 
-def build_scope(scopes: list) -> str:
+def build_scope(scopes: List[AuthScope]) -> str:
     """Builds a valid scope string from list
 
-    :param scopes: list of :class:`~twitchAPI.types.AuthScope`
+    :param list[~twitchAPI.types.AuthScope] scopes: list of :class:`~twitchAPI.types.AuthScope`
     :rtype: str
+    :returns: the valid auth scope string
     """
     return ' '.join([s.value for s in scopes])
 
@@ -114,13 +130,14 @@ def fields_to_enum(data: Union[dict, list],
                    default: Optional[Enum]) -> Union[dict, list]:
     """Itterates a dict or list and tries to replace every dict entry with key in fields with the correct Enum value
 
-    :param data: dict or list
-    :param fields: list of keys to be replaced
+    :param Union[dict, list] data: dict or list
+    :param list[str] fields: list of keys to be replaced
     :param _enum: Type of Enum to be replaced
     :param default: The default value if _enum does not contain the field value
     :rtype: dict or list
     """
     _enum_vals = [e.value for e in _enum.__members__.values()]
+
     def make_dict_field_enum(data: dict,
                              fields: List[str],
                              _enum: Type[Enum],
@@ -143,4 +160,3 @@ def fields_to_enum(data: Union[dict, list],
         return [make_dict_field_enum(d, fields, _enum, default) for d in data]
     else:
         return make_dict_field_enum(data, fields, _enum, default)
-
