@@ -1,4 +1,43 @@
 #  Copyright (c) 2020. Lena "Teekeks" During <info@teawork.de>
+"""
+User OAuth Authenticator and helper functions
+---------------------------------------------
+
+This tool is an alternative to various online services that give you a user auth token.
+
+************
+Requirements
+************
+
+Since this tool opens a browser tab for the Twitch authentication, you can only use this tool on enviroments that can
+open a browser window and render the `<twitch.tv>`__ website.
+
+For my authenticator you have to add the following URL as a "OAuth Redirect URL": :code:`http://localhost:17563`
+You can set that `here in your twitch dev dashboard <https://dev.twitch.tv/console>`__.
+
+************
+Code example
+************
+
+.. code-block:: python
+
+    from twitchAPI.twitch import Twitch
+    from twitchAPI.oauth import UserAuthenticator
+    from twitchAPI.types import AuthScope
+
+    twitch = Twitch('my_app_id', 'my_app_secret')
+
+    target_scope = [AuthScope.BITS_READ]
+    auth = UserAuthenticator(twitch, target_scope, force_verify=False)
+    # this will open your default browser and prompt you with the twitch verification website
+    token, refresh_token = auth.authenticate()
+    # add User authentication
+    twitch.set_user_authentication(token, target_scope, refresh_token)
+
+********************
+Class Documentation:
+********************
+"""
 from .twitch import Twitch
 from .helper import build_url, build_scope, get_uuid, TWITCH_AUTH_BASE_URL
 from .types import AuthScope
@@ -18,11 +57,14 @@ def refresh_access_token(refresh_token: str,
                          app_secret: str):
     """Simple helper function for refreshing a user access token.
 
-    :param refresh_token: str, the current refresh_token
-    :param app_id: str, the id of your app
-    :param app_secret: str, the secret key of your app
+    :param refresh_token: the current refresh_token
+    :type refresh_token: str
+    :param app_id: the id of your app
+    :type app_id: str
+    :param app_secret: the secret key of your app
+    :type app_secret: str
     :return: access_token, refresh_token
-    :rtype: str, str
+    :rtype: (str, str)
     """
     param = {
         'refresh_token': refresh_token,
@@ -39,9 +81,11 @@ def refresh_access_token(refresh_token: str,
 class UserAuthenticator:
     """Simple to use client for the Twitch User authentication flow.
 
-    :param twitch: :class:`twitchAPI.twitch.Twitch` instance
-    :param scopes: List of :class:`twitchAPI.types.AuthScope`, the desired Auth scopes
-    """
+       :param ~twitchAPI.twitch.Twitch twitch: A twitch instance
+       :param list[~twitchAPI.types.AuthScope] scopes: List of the desired Auth scopes
+       :param bool force_verify: If this is true, the user will always be prompted for authorization by twitch,
+                    |default| :code:`False`
+       """
 
     __twitch: 'Twitch' = None
     port: int = 17563
@@ -104,7 +148,7 @@ class UserAuthenticator:
         self.__server_running = True
         try:
             self.__loop.run_until_complete(self.__run_check())
-        except CancelledError:
+        except (CancelledError, asyncio.CancelledError):
             pass
 
     def __start(self):
@@ -143,6 +187,7 @@ class UserAuthenticator:
 
         :param callback_func: Function to call once the authentication finnished.
         :return: None if callback_func is set, otherwise access_token and refresh_token
+        :rtype: None or (str, str)
         """
         self.__callback_func = callback_func
         self.__start()
