@@ -1,4 +1,5 @@
 #  Copyright (c) 2020. Lena "Teekeks" During <info@teawork.de>
+"""Type Definitions"""
 
 from enum import Enum
 
@@ -20,11 +21,15 @@ class AuthScope(Enum):
     :var ANALYTICS_READ_GAMES:
     :var BITS_READ:
     :var CHANNEL_READ_SUBSCRIPTIONS:
+    :var CHANNEL_READ_STREAM_KEY:
+    :var CHANNEL_EDIT_COMMERCIAL:
+    :var CHANNEL_READ_HYPE_TRAIN:
     :var CLIPS_EDIT:
     :var USER_EDIT:
     :var USER_EDIT_BROADCAST:
     :var USER_READ_BROADCAST:
     :var USER_READ_EMAIL:
+    :var USER_EDIT_FOLLOWS:
     :var CHANNEL_MODERATE:
     :var CHAT_EDIT:
     :var CHAT_READ:
@@ -36,11 +41,15 @@ class AuthScope(Enum):
     ANALYTICS_READ_GAMES = 'analytics:read:games'
     BITS_READ = 'bits:read'
     CHANNEL_READ_SUBSCRIPTIONS = 'channel:read:subscriptions'
+    CHANNEL_READ_STREAM_KEY = 'channel:read:stream_key'
+    CHANNEL_EDIT_COMMERCIAL = 'channel:edit:commercial'
+    CHANNEL_READ_HYPE_TRAIN = 'channel:read:hype_train'
     CLIPS_EDIT = 'clips:edit'
     USER_EDIT = 'user:edit'
     USER_EDIT_BROADCAST = 'user:edit:broadcast'
     USER_READ_BROADCAST = 'user:read:broadcast'
     USER_READ_EMAIL = 'user:read:email'
+    USER_EDIT_FOLLOWS = 'user:edit:follows'
     CHANNEL_MODERATE = 'channel:moderate'
     CHAT_EDIT = 'chat:edit'
     CHAT_READ = 'chat:read'
@@ -89,6 +98,19 @@ class SortMethod(Enum):
     VIEWS = 'views'
 
 
+class HypeTrainContributionMethod(Enum):
+    """Enum of valid Hype Train contribution types
+
+    :var BITS:
+    :var SUBS:
+    :var UNKNOWN:
+    """
+
+    BITS = 'BITS'
+    SUBS = 'SUBS'
+    UNKNOWN = ''
+
+
 class VideoType(Enum):
     """Enum of valid video types
 
@@ -99,8 +121,8 @@ class VideoType(Enum):
     :var UNKNOWN:
     """
     ALL = 'all'
-    UPLOAD = 'upload',
-    ARCHIVE = 'archive',
+    UPLOAD = 'upload'
+    ARCHIVE = 'archive'
     HIGHLIGHT = 'highlight'
     UNKNOWN = ''
 
@@ -143,11 +165,26 @@ class CodeStatus(Enum):
     UNKNOWN_VALUE = ''
 
 
-class UnauthorizedException(Exception):
+class TwitchAPIException(Exception):
+    """Base Twitch API Exception"""
+    pass
+
+
+class TwitchAuthorizationException(TwitchAPIException):
+    """Exception in the Twitch Authorization"""
+    pass
+
+
+class UnauthorizedException(TwitchAuthorizationException):
     """Not authorized to use this"""
     pass
 
 
-class MissingScopeException(Exception):
+class MissingScopeException(TwitchAuthorizationException):
     """authorization is missing scope"""
     pass
+
+
+class TwitchBackendException(TwitchAPIException):
+    """when the Twitch API itself is down"""
+    pass
