@@ -1,42 +1,133 @@
 #  Copyright (c) 2020. Lena "Teekeks" During <info@teawork.de>
+"""
+Full Implementation of the Twitch Webhook
+-----------------------------------------
+
+The Webhook runs in its own thread, calling the given callback function whenever an webhook event happens.
+
+Look at the `Twitch Webhook reference <https://dev.twitch.tv/docs/api/webhooks-reference>`__ to find the topics you are
+interested in.
+
+************
+Requirements
+************
+
+You need to have a public IP with a port open. That port will be 80 by default.
+Authentication is off by default but you can choose to authenticate to use some Webhook Topics or to get more information.
+
+.. note:: Please note that Your Endpoint URL has to be HTTPS if you need authentication which means that you probably
+            need a reverse proxy like nginx. This lib currently does not provide any way to add https on its own.
+
+
+*******************
+Short code example:
+*******************
+
+.. code-block:: python
+
+    from twitchAPI.twitch import Twitch
+    from twitchAPI.webhook import TwitchWebHook
+    from pprint import pprint
+
+    def callback_stream_changed(uuid, data):
+        print('Callback for UUID ' + str(uuid))
+        pprint(data)
+
+    twitch = Twitch(td['app_id'], td['secret'])
+    twitch.authenticate_app([])
+
+    user_info = twitch.get_users(logins=['my_twitch_user'])
+    user_id = user_info['data'][0]['id']
+    # basic setup
+    hook = TwitchWebHook("https://my.cool.domain.net:8080", 'my_app_id', 8080)
+    hook.authenticate(twitch)
+    hook.start()
+    print('subscribing to hook:')
+    success, uuid = hook.subscribe_stream_changed(user_id, callback_stream_changed)
+    pprint(success)
+    pprint(twitch.get_webhook_subscriptions())
+    # the webhook is now running and you are subscribed to the topic you want to listen to. lets idle a bit...
+    input('press Enter to shut down...\\n')
+    hook.stop()
+    print('done')
+
+*********************
+Subscription handling
+*********************
+
+You can subscribe to webhook topics using the :code:`subscribe_` prefixed methods.
+
+If :attr:`~.TwitchWebHook.wait_for_subscription_confirm` is True (default), this will wait for the full handshake and
+confirmation to happen, otherwise the returned success value  might be inaccurate in case the subscription itself
+succeeded but the final handshake failed.
+
+You can unsubscribe from a webhook subscription at any time by using :meth:`~twitchAPI.webhook.TwitchWebHook.unsubscribe`
+
+If :attr:`~.TwitchWebHook.unsubscribe_on_stop` is True (default), you dont need to manually unsubscribe from topics.
+
+By deafult, subscriptions will be automatically renewed one minute before they run out for as long as the
+webhook is running.
+
+You can also use :meth:`~twitchAPI.webhook.TwitchWebHook.unsubscribe_all` to unsubscribe from all topic subscriptions at
+once. This will also unsubscribe from topics that where left over from a previous run.
+
+********************
+Class Documentation:
+********************
+"""
+
+
 from typing import Union, Tuple, Callable
-from .helper import build_url, TWITCH_API_BASE_URL, get_uuid, get_json
+from .helper import build_url, TWITCH_API_BASE_URL, get_uuid, get_json, make_fields_datetime, fields_to_enum
+from .helper import extract_uuid_str_from_url
+from .types import *
 import requests
 from aiohttp import web
 import threading
 import asyncio
 from uuid import UUID
-from datetime import datetime
 import logging
-from dateutil import parser as du_parser
+import time
+from .twitch import Twitch
+from concurrent.futures._base import CancelledError
 
 
 class TwitchWebHook:
     """Webhook integration for the Twitch Helix API.
 
-    :param callback_url: The full URL of the webhook.
-    :param api_client_id: The id of your API client
-    :param port: the port on which this webhook should run
-    :var secret: A random secret string. Set this for added security.
-    :var callback_url: The full URL of the webhook.
+    :param str callback_url: The full URL of the webhook.
+    :param str api_client_id: The id of your API client
+    :param int port: the port on which this webhook should run
+    :var str secret: A random secret string. Set this for added security.
+    :var str callback_url: The full URL of the webhook.
+    :var int subscribe_least_seconds: The duration in seconds for how long you want to subscribe to webhhoks.
+                    Min 300 Seconds, Max 864000 Seconds. |default| :code:`600`
+    :var bool auto_renew_subscription: If True, automatically renew all webhooks once they get close to running out.
+                    **Only disable this if you know what you are doing.** |default| :code:`True`
+    :var bool wait_for_subscription_confirm: Set this to false if you dont want to wait for a subscription confirm.
+                    |default| :code:`True`
+    :var int wait_for_subscription_confirm_timeout: Max time in seconds to wait for a subscription confirmation.
+                    Only used if ``wait_for_subscription_confirm`` is set to True. |default| :code:`30`
+    :var bool unsubscribe_on_stop: Unsubscribe all currently active Webhooks on calling `stop()`
+                    |default| :code:`True`
     """
 
     secret = None
-    __client_id = None
-    __auth_token = None
     callback_url = None
-
-    subscribe_least_seconds: int = 864000
-    """The duration in seconds for how long you want to subscribe to webhhoks."""
-
+    subscribe_least_seconds: int = 600
+    auto_renew_subscription: bool = True
+    wait_for_subscription_confirm: bool = True
+    wait_for_subscription_confirm_timeout: int = 30
+    unsubscribe_on_stop: bool = True
     _port: int = 80
     _host: str = '0.0.0.0'
-
+    __twitch: Twitch = None
+    __task_refresh = None
+    __client_id = None
+    __running = False
     __callbacks = {}
-    __urls = {}
-
+    __active_webhooks = {}
     __authenticate: bool = False
-
     __hook_thread: Union['threading.Thread', None] = None
     __hook_loop: Union['asyncio.AbstractEventLoop', None] = None
     __hook_runner: Union['web.AppRunner', None] = None
@@ -46,18 +137,18 @@ class TwitchWebHook:
         self.__client_id = api_client_id
         self._port = port
 
-    def authenticate(self, auth_token: str) -> None:
+    def authenticate(self, twitch: Twitch) -> None:
         """Set authentication for the Webhook. Can be either a app or user token.
 
-        :param auth_token: str
+        :param ~twitchAPI.twitch.Twitch twitch: a authenticated instance of :class:`~twitchAPI.twitch.Twitch`
         :rtype: None
-        :raises: Exception
+        :raises RuntimeError: if the callback URL does not use HTTPS
         """
         self.__authenticate = True
-        self.__auth_token = auth_token
+        self.__twitch = twitch
         if not self.callback_url.startswith('https'):
-            raise Exception('HTTPS is required for authenticated webhook.\n'
-                            + 'Either use non authenticated webhook or use a HTTPS proxy!')
+            raise RuntimeError('HTTPS is required for authenticated webhook.\n'
+                               + 'Either use non authenticated webhook or use a HTTPS proxy!')
 
     def __build_runner(self):
         hook_app = web.Application()
@@ -73,6 +164,8 @@ class TwitchWebHook:
                              web.post('/moderation/moderators/events', self.__handle_moderator_change_events),
                              web.get('/moderation/banned/events', self.__handle_challenge),
                              web.post('/moderation/banned/events', self.__handle_channel_ban_change_events),
+                             web.get('/hypetrain/events', self.__handle_challenge),
+                             web.post('/hypetrain/events', self.__handle_hypetrain_events),
                              web.get('/subscriptions/events', self.__handle_challenge),
                              web.post('/subscriptions/events', self.__handle_subscription_events)])
         hook_runner = web.AppRunner(hook_app)
@@ -85,15 +178,39 @@ class TwitchWebHook:
         self.__hook_loop.run_until_complete(runner.setup())
         site = web.TCPSite(runner, str(self._host), self._port)
         self.__hook_loop.run_until_complete(site.start())
-        print('started twitch API hook on port ' + str(self._port))
-        self.__hook_loop.run_forever()
+        logging.info('started twitch API hook on port ' + str(self._port))
+        # add refresh task
+        if self.auto_renew_subscription:
+            self.__task_refresh = self.__hook_loop.create_task(self.__refresh_task())
+        try:
+            self.__hook_loop.run_forever()
+        except (CancelledError, asyncio.CancelledError):
+            pass
+
+    async def __refresh_task(self):
+        while True:
+            # renew 1 Min before timer runs out:
+            await asyncio.sleep(self.subscribe_least_seconds - 60)
+            # make sure that the auth token is still valid:
+            if self.__authenticate:
+                self.__twitch.refresh_used_token()
+            for key in self.__active_webhooks.keys():
+                self.renew_subscription(key)
 
     def start(self):
         """Starts the Webhook
 
         :rtype: None
+        :raises ValueError: if subscribe_least_seconds is not in range 300 to 864000
+        :raises RuntimeError: if webhook is already running
         """
+        if self.subscribe_least_seconds < 60 * 5 or self.subscribe_least_seconds > 864000:
+            # at least 5 min, max 864000 seconds
+            raise ValueError('subscribe_least_second has to be in range 300 to 864000')
+        if self.__running:
+            raise RuntimeError('already started')
         self.__hook_thread = threading.Thread(target=self.__run_hook, args=(self.__build_runner(),))
+        self.__running = True
         self.__hook_thread.start()
 
     def stop(self):
@@ -104,9 +221,16 @@ class TwitchWebHook:
         :rtype: None
         """
         if self.__hook_runner is not None:
+            if self.unsubscribe_on_stop:
+                all_keys = list(self.__active_webhooks.keys())
+                for uuid in all_keys:
+                    self.unsubscribe(uuid)
+            if self.auto_renew_subscription:
+                self.__task_refresh.cancel()
             self.__hook_loop.call_soon_threadsafe(self.__hook_loop.stop)
             self.__hook_runner = None
             self.__hook_thread.join()
+            self.__running = False
 
     # ==================================================================================================================
     # HELPER
@@ -117,7 +241,10 @@ class TwitchWebHook:
             "Client-ID": self.__client_id
         }
         if self.__authenticate:
-            headers['Authorization'] = "Bearer " + self.__auth_token
+            token = self.__twitch.get_used_token()
+            if token is None:
+                raise TwitchAuthorizationException('no Authorization set!')
+            headers['Authorization'] = "Bearer " + token
         return headers
 
     def __api_post_request(self, url: str, data: Union[dict, None] = None):
@@ -139,12 +266,14 @@ class TwitchWebHook:
             arr.append(callback_func)
         self.__callbacks[uuid] = arr
 
-    def _subscribe(self, callback_path: str, topic_url: str, mode: str = "subscribe"):
+    def _subscribe(self, callback_path: str, topic_url: str, mode: str = "subscribe", callback_full=True):
         """"Subscribe to Twitch Topic"""
         data = {'hub.callback': self.callback_url + callback_path,
                 'hub.mode': mode,
                 'hub.topic': topic_url,
                 'hub.lease_seconds': self.subscribe_least_seconds}
+        if not callback_full:
+            data['hub.callback'] = callback_path
         if self.secret is not None:
             data['hub.secret'] = self.secret
         result = self.__api_post_request(TWITCH_API_BASE_URL + "webhooks/hub", data=data)
@@ -156,18 +285,24 @@ class TwitchWebHook:
         success = self._subscribe(callback_path+"?uuid=" + str(uuid), url)
         if success:
             self.__add_callable(uuid, callback_func)
-            self.__urls[uuid] = url
+            # self.__urls[uuid] = url
+            self.__active_webhooks[uuid] = {
+                'url': url,
+                'callback': callback_func,
+                'callback_path': callback_path + "?uuid=" + str(uuid),
+                'confirmed_subscribe': False,
+                'confirmed_unsubscribe': False,
+                'active': False
+            }
+            if self.wait_for_subscription_confirm:
+                timeout = time.time() + self.wait_for_subscription_confirm_timeout
+                while timeout > time.time() and not self.__active_webhooks.get(uuid)['confirmed_subscribe']:
+                    time.sleep(0.1)
+                return self.__active_webhooks.get(uuid)['confirmed_subscribe']
         return success
 
-    def _generic_unsubscribe(self, callback_path: str, uuid: UUID) -> bool:
-        url = self.__urls.get(uuid)
-        if url is None:
-            raise Exception(f'no subscription found for UUID {str(uuid)}')
-        success = self._subscribe(callback_path + "?uuid=" + str(uuid), url, mode="unsubscribe")
-        if success:
-            self.__urls.pop(uuid, None)
-            self.__callbacks.pop(uuid, None)
-        return success
+    def _generic_unsubscribe(self, callback_path: str, url: str, callback_full: bool = True) -> bool:
+        return self._subscribe(callback_path, url, mode="unsubscribe", callback_full=callback_full)
 
     def _generic_handle_callback(self, request: 'web.Request', data: Union[dict, list, None]) -> 'web.Response':
         uuid_str = request.rel_url.query.get('uuid')
@@ -181,6 +316,73 @@ class TwitchWebHook:
             cf(uuid, data)
         return web.Response(text="")
     # ==================================================================================================================
+    # SUBSCRIPTION HELPER
+    # ==================================================================================================================
+
+    __unsubscribe_all_helper = {}
+
+    def unsubscribe_all(self,
+                        twitch: Twitch) -> bool:
+        """Unsubscribe from all Webhooks that use the callback URL set in `callback_url`\n
+        **If `wait_for_subscription_confirm` is False, the response might be
+        True even tho the unsubscribe action failed.**
+
+        :param ~twitchAPI.twitch.Twitch twitch: App authorized instance of :class:`~twitchAPI.twitch.Twitch`
+        :rtype: bool
+        :returns: True if all webhooks could be unsubscribed, otherwise False.
+        """
+        self.__unsubscribe_all_helper = {}
+        data = twitch.get_webhook_subscriptions()
+        sub_responses = []
+        for d in data.get('data', []):
+            uuid = extract_uuid_str_from_url(d.get('callback'))
+            if uuid is not None and d.get('callback').startswith(self.callback_url):
+                self.__unsubscribe_all_helper[uuid] = False
+                sub_responses.append(self._generic_unsubscribe(d.get('callback'), d.get('topic'), callback_full=False))
+        if self.wait_for_subscription_confirm:
+            timeout = time.time() + self.wait_for_subscription_confirm_timeout
+            while timeout > time.time() and not all(self.__unsubscribe_all_helper.values()):
+                time.sleep(0.1)
+            return all(self.__unsubscribe_all_helper.values()) and all(sub_responses)
+        else:
+            return all(sub_responses)
+
+    def renew_subscription(self,
+                           uuid: UUID) -> bool:
+        """Renew existing topic subscription
+
+        :param uuid: UUID of the subscription to renew
+        :rtype: bool
+        :returns: True if renewal worked. Note that you still need to wait for the handshake to make sure its renewed.
+        """
+        url = self.__active_webhooks.get(uuid)
+        if url is None:
+            raise Exception(f'no subscription found for UUID {str(uuid)}')
+        logging.info('renewing webhook ' + str(uuid))
+        return self._subscribe(url.get('callback_path'), url.get('url'))
+
+    def unsubscribe(self,
+                    uuid: UUID) -> bool:
+        url = self.__active_webhooks.get(uuid)
+        if url is None:
+            raise Exception(f'no subscription found for UUID {str(uuid)}')
+        success = self._generic_unsubscribe(url.get('callback_path'), url.get('url'))
+        if success:
+            self.__callbacks.pop(uuid, None)
+            if self.wait_for_subscription_confirm:
+                timeout = time.time() + self.wait_for_subscription_confirm_timeout
+                while timeout > time.time() and not self.__active_webhooks.get(uuid)['confirmed_unsubscribe']:
+                    time.sleep(0.05)
+                if self.__active_webhooks.get(uuid)['confirmed_unsubscribe']:
+                    self.__active_webhooks.pop(uuid)
+                else:
+                    # unsubscribe failed!
+                    return False
+
+
+        return success
+
+    # ==================================================================================================================
     # SUBSCRIPTIONS
     # ==================================================================================================================
 
@@ -207,14 +409,6 @@ class TwitchWebHook:
         uuid = get_uuid()
         return self._generic_subscribe('/users/follows', url, uuid, callback_func), uuid
 
-    def unsubscribe_user_follow(self, uuid: UUID) -> bool:
-        """Unsubscribe from user follow topic
-
-        :param uuid: UUID of the subscription
-        :rtype: bool
-        """
-        return self._generic_unsubscribe('/users/follows', uuid)
-
     def subscribe_stream_changed(self,
                                  user_id: str,
                                  callback_func: Union[Callable[[UUID, dict], None], None]) -> Tuple[bool, UUID]:
@@ -230,14 +424,6 @@ class TwitchWebHook:
         uuid = get_uuid()
         return self._generic_subscribe('/streams', url, uuid, callback_func), uuid
 
-    def unsubscribe_stream_changed(self, uuid: UUID) -> bool:
-        """Unsubscribe from stream changed topic
-
-        :param uuid: UUID of the subscription
-        :return: bool
-        """
-        return self._generic_unsubscribe('/streams', uuid)
-
     def subscribe_user_changed(self,
                                user_id: str,
                                callback_func: Union[Callable[[UUID, dict], None], None]) -> Tuple[bool, UUID]:
@@ -253,14 +439,6 @@ class TwitchWebHook:
         uuid = get_uuid()
         return self._generic_subscribe('/users/changed', url, uuid, callback_func), uuid
 
-    def unsubscribe_user_changed(self, uuid: UUID) -> bool:
-        """Unsubscribe from subscription event topic
-
-        :param uuid: UUID of the subscription
-        :rtype: bool
-        """
-        return self._generic_unsubscribe("/users/changed", uuid)
-
     def subscribe_extension_transaction_created(self,
                                                 extension_id: str,
                                                 callback_func: Union[Callable[[UUID, dict], None], None]) \
@@ -283,14 +461,6 @@ class TwitchWebHook:
         uuid = get_uuid()
         return self._generic_subscribe('/extensions/transactions', url, uuid, callback_func), uuid
 
-    def unsubscribe_extension_transactions_created(self, uuid: UUID) -> bool:
-        """Unsubscribe from Extension transaction created Topic
-
-        :param uuid: UUID of the subscription
-        :rtype: bool
-        """
-        return self._generic_unsubscribe('/extensions/transactions', uuid)
-
     def subscribe_moderator_change_events(self,
                                           broadcaster_id: str,
                                           user_id: Union[str, None],
@@ -312,14 +482,6 @@ class TwitchWebHook:
         uuid = get_uuid()
         return self._generic_subscribe('/moderation/moderators/events', url, uuid, callback_func), uuid
 
-    def unsubscribe_moderator_change_events(self, uuid: UUID) -> bool:
-        """Unsubscribe from Moderator Change Events Topic
-
-        :param uuid: UUID of the subscription
-        :rtype: bool
-        """
-        return self._generic_unsubscribe('/moderation/moderators/events', uuid)
-
     def subscribe_channel_ban_change_events(self,
                                             broadcaster_id: str,
                                             user_id: Union[str, None],
@@ -341,14 +503,6 @@ class TwitchWebHook:
         uuid = get_uuid()
         return self._generic_subscribe('/moderation/banned/events', url, uuid, callback_func), uuid
 
-    def unsubscribe_channel_ban_change_events(self, uuid: UUID) -> bool:
-        """Unsubscribe from Channel Ban Change Events Topic
-
-        :param uuid: UUID of the subscription
-        :rtype: bool
-        """
-        return self._generic_unsubscribe('/moderation/banned/events', uuid)
-
     def subscribe_subscription_events(self,
                                       broadcaster_id: str,
                                       callback_func: Union[Callable[[UUID, dict], None]],
@@ -376,13 +530,23 @@ class TwitchWebHook:
         uuid = get_uuid()
         return self._generic_subscribe('/subscriptions/events', url, uuid, callback_func), uuid
 
-    def unsubscribe_subscription_events(self, uuid: UUID) -> bool:
-        """Unsubscribe from Subscription Events Topic
+    def subscribe_hype_train_events(self,
+                                    broadcaster_id: str,
+                                    callback_func: Union[Callable[[UUID, dict], None]]) -> Tuple[bool, UUID]:
+        """Subscribe to Hype Train Events\n
+        See https://dev.twitch.tv/docs/api/webhooks-reference#topic-hype-train-event for documentation
 
-        :param uuid: UUID of the subscription
-        :rtype: bool
+        :param broadcaster_id: str
+        :param callback_func: function for callback
+        :rtype: bool, UUID
         """
-        return self._generic_unsubscribe('/subscriptions/events', uuid)
+        params = {
+            'broadcaster_id': broadcaster_id,
+            'first': 1
+        }
+        url = build_url(TWITCH_API_BASE_URL + 'hypetrain/events', params)
+        uuid = get_uuid()
+        return self._generic_subscribe('/hypetrain/events', url, uuid, callback_func), uuid
 
     # ==================================================================================================================
     # HANDLERS
@@ -397,7 +561,7 @@ class TwitchWebHook:
         if d is not None:
             if len(d['data']) > 0:
                 data = d['data'][0]
-                data['started_at'] = datetime.fromisoformat(data['started_at'])
+                data = make_fields_datetime(data, ['started_at'])
             else:
                 data = {
                     'type': 'offline'
@@ -408,7 +572,7 @@ class TwitchWebHook:
         data = await get_json(request)
         if data is not None:
             data = data['data'][0]
-            data['followed_at'] = du_parser.isoparse(data['followed_at'])
+            data = make_fields_datetime(data, ['followed_at'])
         return self._generic_handle_callback(request, data)
 
     async def __handle_user_changed(self, request: 'web.Request'):
@@ -422,13 +586,28 @@ class TwitchWebHook:
         data = d
         if data is not None:
             data = data['data'][0]
-            data['timestamp'] = du_parser.isoparse(data['timestamp'])
+            data = make_fields_datetime(data, ['timestamp'])
         return self._generic_handle_callback(request, data)
 
     async def __handle_challenge(self, request: 'web.Request'):
         challenge = request.rel_url.query.get('hub.challenge')
         if challenge is not None:
             # found challenge, lets answer it
+            if request.rel_url.query.get('hub.mode') == 'subscribe':
+                # we treat this as active as soon as we answer the challenge
+                self.__active_webhooks.get(UUID(request.rel_url.query.get('uuid')))['active'] = True
+                self.__active_webhooks.get(UUID(request.rel_url.query.get('uuid')))['confirmed_subscribe'] = True
+            if request.rel_url.query.get('hub.mode') == 'unsubscribe':
+                uuid_str = request.rel_url.query.get('uuid')
+                if uuid_str in self.__unsubscribe_all_helper.keys():
+                    self.__unsubscribe_all_helper[uuid_str] = True
+                if UUID(uuid_str) in self.__active_webhooks.keys():
+                    # we treat this as invalid as soon as we answer the challenge
+                    if self.wait_for_subscription_confirm:
+                        self.__active_webhooks.get(UUID(request.rel_url.query.get('uuid')))['confirmed_unsubscribe'] = True
+                    else:
+                        self.__active_webhooks.pop(UUID(request.rel_url.query.get('uuid')))
+
             return web.Response(text=challenge)
         return web.Response(status=500)
 
@@ -436,19 +615,30 @@ class TwitchWebHook:
         data = await get_json(request)
         if data is not None:
             data = data['data'][0]
-            data['event_timestamp'] = du_parser.isoparse(data['event_timestamp'])
+            data = make_fields_datetime(data, ['event_timestamp'])
             return self._generic_handle_callback(request, data)
 
     async def __handle_channel_ban_change_events(self, request: 'web.Request'):
         data = await get_json(request)
         if data is not None:
             data = data['data'][0]
-            data['event_timestamp'] = du_parser.isoparse(data['event_timestamp'])
+            data = make_fields_datetime(data, ['event_timestamp'])
         return self._generic_handle_callback(request, data)
 
     async def __handle_subscription_events(self, request: 'web.Request'):
         data = await get_json(request)
         if data is not None:
             data = data['data'][0]
-            data['event_timestamp'] = du_parser.isoparse(data['event_timestamp'])
+            data = make_fields_datetime(data, 'event_timestamp')
+        return self._generic_handle_callback(request, data)
+
+    async def __handle_hypetrain_events(self, request: 'web.Request'):
+        data = await get_json(request)
+        if data is not None:
+            data = data['data'][0]
+            data = make_fields_datetime(data, ['event_timestamp',
+                                               'cooldown_end_time',
+                                               'expires_at',
+                                               'started_at'])
+            data = fields_to_enum(data, ['type'], HypeTrainContributionMethod, HypeTrainContributionMethod.UNKNOWN)
         return self._generic_handle_callback(request, data)
